; Scripts for Youtube.
;
; Author: Doug Lee
/*
Copyright (c) 2016-2018 Doug Lee.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* The names of the copyright holders and contributors may not be used to
  endorse or promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

include "hjconst.jsh"
include "youtube.jsh"
include "youtube.jsm"

const scriptVer = 41

script scriptFileName()
scriptAndAppNames(formatString(msgScriptName, scriptVer))
endScript

void function autoStartEvent()
apphelp__setDocName("youtube.htm")
mkcs__setCommandKey("[")
mkcs__setHelpText(msgMKCSHelpText)
endFunction

script skipAd()
; Skip a playing video ad.
if !fsxml__performAction(Action_DoDefaultAction, searchSkipAd)
	sayMessage(OT_Error, msgNoSkipAd)
	return
endIf
endScript

script pressPlayPause()
; Press the Play/Pause button if present.
if !fsxml__performAction(Action_DoDefaultAction, searchPlayPause)
	sayMessage(OT_Error, msgNoPlayPause)
	return
endIf
endScript

script focusVolumeSlider()
; Set focus to the volume slider of a video, from which all keys work.
if !fsxml__performAction(Action_SetFocus, searchVolume)
	sayMessage(OT_Error, msgNoVolume)
	return
endIf
routeVirtualToPC()
if !isFormsModeActive() && isVirtualPCCursor()
	turnOnFormsMode()
	sayObjectTypeAndText()
endIf
endScript

script readTime()
; Reads the time played and total and also tries to make them stay up to date.
var object doc = fsxml__virtualCursorContext(-3)
var object o = doc.selectSingleNode("//node()[@class='ytp-time-display notranslate']")
if !o
	beep()
	return
endIf
var string fsid = o.getAttribute("fsID")
; This should make the video time appear and stay up to date.
; This is disabled because it fails to cause continuous time updates and also causes side effects like focus changes sometimes.
; [DGL, 2016-07-23]
;tryToMoveMouseToElement(fsid)
;pause()
sayMessage(OT_User_Requested_Information, o.text)
endScript

script GoToOwningChannel()
; :Synopsis Goes to the channel that owns the current video.
; This script goes to the channel that owns the video that is currently playing. Does nothing if you are not on a video page.
if !fsxml__performAction(Action_DoDefaultAction, searchVideoOwner)
sayMessage (OT_Error, msgNoOwner)
	return
endIf
endScript

int function tryToMoveMouseToElement(string fsid)
; Try to move the mouse to the center of the given element.
; First attempts to make the element visible so it will have a valid rectangle.
; Returns True on apparent success and False on failure to make the element visible.
var collection rect = fsxml__performAction(Action_MakeVisible, fsid)
if rect
	var int restore = False
	if !isJAWSCursor()
		saveCursor()
		JAWSCursor()
		restore = True
	endIf
	moveTo( (rect.left+rect.right)/2, (rect.top+rect.bottom)/2)
	if restore
		restoreCursor()
	endIf
else
	sayMessage(OT_Error, msgNoMakeVisible)
	return False
endIf
return True
endFunction

script toggleMute()
; Toggle the muted status of the currently playing video.
sayMessage(OT_Message, getMuteName())
if !fsxml__performAction(Action_DoDefaultAction, searchMute)
	sayMessage(OT_Error, msgNoNext)
	return
endIf
endScript

string function getMuteName()
; Get the name of the Mute button, which names the action that clicking it will perform.
; Because the actual button is sometimes not named usefully, this function uses the name of the Volume panel as a backup to figure it out.
; Returns the null string on error or failure.
var object doc = fsxml__virtualCursorContext(-3, True)
var object o = doc.selectSingleNode("//node()[@class='ytp-mute-button ytp-button']")
if !o
	return ""
endIf
var string name = o.getAttribute("fsText")
if !stringIsBlank(name) && !stringStartsWith(name, scUnlabeled)
	return name
endIf
; Plan B...
o = doc.selectSingleNode("//node()[@class='ytp-volume-panel']")
if !o
	return ""
endIf
name = o.getAttribute("fsText")
name = string__segmentNextFromRight(name, " ")
if name == scMuted
	return msgUnmute
elif name == scVolume
	return msgMute
endIf
return ""
endFunction

script clickNext()
; Click the first Next button or link found on the current page.
if !fsxml__performAction(Action_DoDefaultAction, searchNext)
&& !doNav(1)
	sayMessage(OT_Error, msgNoNext)
	return
endIf
endScript

script clickPrevious()
; Click the first Previous button or link found on the current page.
if !fsxml__performAction(Action_DoDefaultAction, searchPrevious)
&& !doNav(-1)
	sayMessage(OT_Error, msgNoPrevious)
	return
endIf
endScript

int function doNav(int delta)
var string pclass = "branded-page-box search-pager  spf-link "
; Warning: The space at the end of the above class value matters!
var object doc = fsxml__virtualCursorContext(-3)
if !doc
	;sayString("noDoc")
	return
endIf
var object o = doc.selectSingleNode("//node()[@class='" +pclass +"']")
; That parents the nav links to go to pages relative to the current one.
o = o.firstChild
if !o
	;sayString("noParentOrChild")
	return
endIf
; This should be the Previous link, hence one page number below current.
while o && o.getAttribute("fsType") != WT_Button
	o = o.nextSibling
endWhile
if o.getAttribute("fsType") != WT_Button
	;sayString("noCurrent")
	return
endIf
while delta && o
	if delta > 0
		o = o.nextSibling
		delta = delta -1
	else
		o = o.previousSibling
		delta = delta +1
	endIf
endWhile
if !o || o.getAttribute("fsType") != WT_Link
	;sayString("noLink")
	return
endIf
var string fsid = o.getAttribute("fsID")
if !fsid
	;sayString("noFSID")
	return
endIf
sayMessage(OT_User_Requested_Information, o.getAttribute("fsText"))
fsxml__performAction(Action_DoDefaultAction, fsid)
return True
endFunction


;================================= Module fsxml ================================
; Utility code for the fsXML support added in JAWS 17 for scripting against web pages.
;
; Author: Doug Lee

collection function fsxml__performAction(int action, variant spec)
; Perform one or more searches to find an element, and apply the given action to it.
; On success, returns a collection indicating the element's rectangle (left/right/top/bottom collection members).
; Other values may also be present; see below.
; On failure, returns null. A boolean context is fine when the rectangle is not needed.
; Allowed values/formats of spec:
;	MSXML object: The fsID attribute value is used.
;	XML string: The LAST fsID is used. This makes passing getElementXML() work (JAWS 17.0.1806+, April 2016).
;	fsID (hex): The FSID given is used.
;	One or more tag-and-attribute search lines: See below.
; For all of the above formats except tag-and-attribute searches, fsid (hex) is included as a returning collection member.
; For tag-and-attribute searches, the following rules apply:
;	- One search specification per line. This makes message constants an ideal way to list search criteria.
;	  When there are multiple lines, the first match wins.
;	- The first character on each line is the line's field delimiter, such as |.
;	- A delimiter may be anything except \r, \n, <, or an alphanumeric character.
;	- Line format (using | for delimiter here): |tagName|attribName|attribValue.
; When a tag-and-attribute search succeeds, the returned collection includes the following extra members (and not fsid):
;	tagName, attribName, attribValue: The values that found the element.
;	searchIndex: The 1-based index of those values among lines of spec.
;	searchSpec: The actual spec line that found the element.
var collection result result = new collection
var int left, int right, int top, int bottom, int success = False
var string fsid = spec.getAttribute("fsID")
if fsid
	; An MSXML object.
	success = performActionOnElementWithID(action, hexToDec(fsid), left, top, right, bottom)
	result.fsid = fsid
elif !stringLength(spec)
	; Anything but a string from now on is an error.
	return
elif stringStartsWith(spec, "<")
	; An XML string.
	var int pos = stringContainsFromRight(spec, "fsID=\"")
	if !pos
		; Probably an XML string that is not fsXML.
		return
	endIf
	spec = stringChopLeft(spec, pos+5)
	fsid = stringSegment(spec, "\"", 1)
	success = performActionOnElementWithID(action, hexToDec(fsid), left, top, right, bottom)
	result.fsid = fsid
elif stringContains("0123456789abcdefABCDEF", stringLeft(spec, 1))
&& stringLength(spec) <= 8
	; An fsID in hex. These should be in lower case but upper is allowed here to be safe.
	success = performActionOnElementWithID(action, hexToDec(spec), left, top, right, bottom)
	result.fsid = spec
else
	; All we have left are tag-and-attribute searches.
	var string lsep = "\n"
	var int n = stringSegmentCount(spec, lsep)
	var int i = 1
	while i <= n && !success
		var string seg = stringSegment(spec, lsep, i)
		var string fsep = stringLeft(seg, 1)
		; First field is null, so we start at 2.
		var string tagName = stringSegment(seg, fsep, 2)
		var string attribName = stringSegment(seg, fsep, 3)
		var string attribValue = stringSegment(seg, fsep, 4)
		success = performActionOnElementWithTagAndAttribute(action, tagName, attribName, attribValue, left, top, right, bottom)
		if success
			result.tagName = tagName
			result.attribName = attribName
			result.attribValue = attribValue
			result.searchIndex = i
			result.searchSpec = seg
		endIf
		i = i +1
	endWhile
endIf
if !success
	return
endIf
result.left = left
result.right = right
result.top = top
result.bottom = bottom
return result
endFunction

object function fsxml__virtualCursorContext(int howMuch, optional int silent, int preserveWhitespace)
; Obtain an MSXML object and possibly containing tree based on the virtual PC cursor's current location.
; Returns the object representing the cursor location or the tree root as appropriate, or null on error.
; howMuch: 0 for just the current object, 1 for its parent also, 2 for its grandparent, etc. -1 includes all ancestors.
; The JAWS getElementXML() function works the same way. Extra values allowed here but not there:
;	-2: Include the entire tree surrounding the current location.
;	-3: Return the entire tree by its root instead of by the current location (equivalent to getXMLDocument()).
; -3 is supported by all JAWS 17 versions; all other values of howMuch require JAWS 17.0.1806 or later.
; If silent is not passed and True, error messages speak as appropriate.  This includes parse errors from oDoc.loadXML().
; Whitespace in object .xml properties is preserved if preserveWhitespace is passed and True, otherwise not.
var string fsid = ""
if howMuch > -3
	; All howMuch values above -3 require that we find the cursor's specific element, for which we need its id.
	; Even getElementXML(0) returns a multi-object tree.
	fsid = fsxml__getElementFSID(silent)
	if !fsid
		return
	endIf
endIf
; fsid is now non-null if and only if we need to use it.

; Get the XML for the tree actually requested.
var string func = "GetDocumentXML"  ; used for error messages
var string xml
if howMuch >= -1
	; Just ancestry, not entire tree wanted.
	xml = getElementXML(howMuch)
	func = "GetElementXML"
elif howMuch == -2 || howMuch == -3
	; Tree surrounding cursor or tree by its root.
	xml = getDocumentXML()
else
	; Invalid howMuch value.
	if !silent
		sayMessage(OT_Error, formatString(fsxml___msgInvalidHowMuch, "fsxml__virtualCursorContext", intToString(howMuch)))
	endIf
	return
endIf
if !xml
	if !silent
		sayMessage(OT_Error, formatString(fsxml___msgNoXML, func))
	endIf
	return
endIf

; Convert the XML to an object or tree. This will get us the root of that tree.
var object o = util__XMLDocument(xml, silent, preserveWhitespace)
if !o
	; util__XMLDocument() spoke any error messages if appropriate.
	return
endIf

; If we need the cursor object rather than the root object, find it now.
if fsid
	o = o.selectSingleNode("//node()[@fsID='" +fsid +"']")
	if !o
		if !silent
			sayMessage(OT_Error, fsxml___msgNoSelect)
		endIf
		return
	endIf
endIf
return o
endFunction

string function fsxml__getElementFSID(optional int silent)
; Return the hex ID of the element at the current virtual cursor location. Returns null on error.
; If silent is not passed and True, error messages speak as appropriate.  This includes parse errors from oDoc.loadXML().
var string xml = getElementXML()
if !xml
	if !silent
		sayMessage(OT_Error, formatString(fsxml___msgNoXML, "getElementXML"))
		sayMessage(OT_Error, fsxml___msgNoID)
	endIf
	return ""
endIf
var int pos = stringContainsFromRight(xml, "fsID=\"")
if !pos
	if !silent
		sayMessage(OT_Error, fsxml___msgNoID)
	endIf
	return ""
endIf
xml = stringChopLeft(xml, pos+5)
xml = stringSegment(xml, "\"", 1)
if !xml
	if !silent
		sayMessage(OT_Error, fsxml___msgNoID)
	endIf
	return ""
endIf
return xml
endFunction

;=============================== End module fsxml ==============================


;================================= Module mkcs =================================
; This module implements a KeymapChangedEvent-based multi-key command sequence system.
;
; Usage:
; Include this module in the main jss.
; Create a message constant for the MKCS help text (see the mkcs__help script for format info).
; Load the help text from AutoStartEvent with a call to mkcs__setHelpText().
; Set the command key (and again if it changes) by calling mkcs__setCommandKey().
; Write key mappings in the jkm file using "&" between layers on each line.
; For each command level, make the following additional mappings (the examples are for the top level with [ as the command key):
;	[&Escape=mkcs__exit
;	[&Tab*=mkcs__help("next")
;	[&Shift+Tab*=mkcs__help("prev")
;	[&JAWSKey+Tab*=mkcs__help("cur")
;	[&Insert+Tab*=mkcs__help("cur")
;	[&Enter=mkcs__help("exec")
; NOTE: The first line above is mandatory, as it is also used by this code
; to find the original command key when changing it to a new one.
; See mkcs__setCommandKey().
;
; Author:  Doug Lee
;
; Revision History:  See version control logs.

include "hjglobal.jsh"

; For support of JAWS versions before 14.0.
prototype string function getScriptSynopsis(string scriptName, optional int useDefault)
prototype string function getScriptDescription(string scriptName, optional int useDefault)
; And for JAWS versions before 17.0...
prototype string function getActiveConfiguration(optional int wantDomain)

globals
; The path to the sound file to play on exit from a layer.
	string mkcs___soundPath,
; For default-level scripts, the namespace for the CommandKey value in the MKCSOptions section of the effective config file.
	string mkcs___namespace,
; The file name and jcf section used by the scripts, if explicitly set in a mkcs__setCommandKey call.
	string mkcs___fname,
	string mkcs___sect,
; The command key that starts all sequences recognized here.
	string mkcs___commandKey

string function mkcs___getActiveConfiguration(optional int wantDomain)
; Wrapper for the JAWS getActiveConfiguration() function that returns the application config when True is passed but no domain scripts are active.
; The JAWS documentation says getActiveConfiguration does this, but through at least JAWS 18 it returns "" in this case. [DGL, 2017-08-19
if !wantDomain
	return getActiveConfiguration()
endIf
var string cfg = getActiveConfiguration(wantDomain)
if cfg
; Logic: If there's no script by this name, this script is not named after the domain and would probably not expect its configuration to be either.
&& fileExists(getJAWSSettingsDirectory() +"\\" +cfg +".jsb")
	return cfg
endIf
return getActiveConfiguration()
endFunction

script mkcs__editCommandKey()
; Allow the user to alter the command key used to start key sequences for these scripts.
; Traditionally assigned to JAWSKey+C / Insert+C at the first level after the currently active command key.
var string sKey = mkcs__getCommandKey()
var string sKey0 = sKey
var int okPressed = inputBox(mkcs___msgEditPrompt, mkcs___msgEditTitle, sKey)
if !okPressed || stringCompare(sKey, sKey0, True) == 0
	sayMessage(OT_User_Requested_Information, mkcs___msgNoChange)
	return
endIf

; Set the key in the active or appropriate config file first, so mkcs__setCommandKey() will find it correctly.
var string jcfname = mkcs___fname
; That would be a jkm file name if not blank.
if jcfname
jcfname = stringLeft(jcfname, stringContainsFromRight(jcfname, ".")-1) +".jcf"
else
	jcfname = mkcs___getActiveConfiguration(True) +".jcf"
endIf
if !stringContains(jcfname, "\\")
	jcfname = getJAWSSettingsDirectory() +"\\" +jcfname
endIf
var string jcfSect = "MKCSOptions"
var string jcfKey = mkcs___namespace +"CommandKey"
var string userCommandKey = iniReadString(jcfSect, jcfKey, "", jcfname)
if stringCompare(userCommandKey, sKey) == 0
	sayMessage(OT_User_Requested_Information, mkcs___msgNoChange)
	return
endIf
iniWriteString(jcfSect, jcfKey, sKey, jcfname)
mkcs__setCommandKey(sKey)
endScript

int function mkcs__setCommandKey(string sKey, /*?*/ string fname, string sect)
; Sets the key used as a prefix for all command sequences in this application.
; Also performs other minor initialization tasks.
; Call from AutoStartEvent or appropriate initialization code.
; Also call if the key is ever changed.
; Updates the jkm file if necessary.
; If fname is not passed, defaults to <activeConfiguration>.jkm.
; If sect is not passed, defaults to "Common Keys."
; The given command key can be overridden by a value for CommandKey in an MKCSOptions section of the config file as of June 13, 2017.
; When using this in default-level scripts, first call mkcs__setNamespace("MyScripts") so the value will come from MyScriptsCommandKey instead.
; The config file defaults to <activeConfiguration>.jcf but gets the basename from fname if passed.
; Returns 0 if things are not left in sync, 1 if they are, and 2 if the jkm changed.
; Things should not be left out of sync unless there is an error.
mkcs___soundPath = ""
if !sKey
	; Can't set to null key, doing nothing is inconsistent,
	; so we set to a string that is not a valid key.
	sKey = "<NoKey>"
endIf
mkcs___commandKey = sKey
if fname
	mkcs___fname = fname
else
	fname = mkcs___fname
endIf
if !fname
	fname = mkcs___getActiveConfiguration(True) +".jkm"
endIf
if !stringContains(fname, "\\")
	fname = getJAWSSettingsDirectory() +"\\" +fname
endIf
if sect
	mkcs___sect = sect
else
	sect = mkcs___sect
endIf
if !sect
	sect = "Common Keys"
endIf
; Allow an MKCSOptions config value to override the command key.
; This is mainly for non-English installations, where author command key choice may block standard things like accented letter typing.
var string jcfname = stringLeft(fname, stringContainsFromRight(fname, ".")-1) +".jcf"
var string jcfSect = "MKCSOptions"
var string jcfKey = mkcs___namespace +"CommandKey"
var string userCommandKey = iniReadString(jcfSect, jcfKey, "", jcfname)
if userCommandKey
	mkcs___commandKey = userCommandKey
endIf

var string jkm = util__stringFromFile(fname)
if !jkm
	; Chances are this means we're on the secure desktop and can't read files.
	return 0
endIf
var string sep = "\n"
var int pos = stringContains(stringLower(jkm), stringLower("&escape=mkcs__exit"))
if !pos
	sayMessage(OT_Error, mkcs___msgNotFound1)
	return 0
endIf
var string sKey0 = stringLeft(jkm, pos-1)
pos = stringContainsFromRight(sKey0, sep)
if !pos
	sayMessage(OT_Error, mkcs___msgNotFound2)
	return 0
endIf
sKey0 = stringChopLeft(sKey0, pos)
; In case the first layer entry is not the command key entry.
pos = stringContainsFromRight(sKey0, "&")
if pos
	sKey0 = stringLeft(sKey0, pos-1)
endIf
if stringCompare(sKey0, mkcs___commandKey, False) == 0
	return 1
endIf

sayUsingVoice(VCTX_Message, mkcs___msgUpdating, OT_Message)
var int n = stringSegmentCount(jkm, sep)
var string jkm1 = ""
var int i = 1
var int changed = False
var int inSect = False
var string line
while i <= n
	line = stringSegment(jkm, sep, i)
	line = stringTrimLeadingBlanks(stringTrimTrailingBlanks(line))
	if stringStartsWith(line, "[" +sect +"]")
		inSect = True
	elif stringLeft(line, 1) == "[" && stringRight(line, 1) == "]"
		inSect = False
	elif inSect
		if stringStartsWith(line, sKey0+"&")
			line = mkcs___commandKey +stringChopLeft(line, stringLength(sKey0))
			changed = True
		endIf
	endIf
	if jkm1
		jkm1 = jkm1 +"\r" +sep
	endIf
	jkm1 = jkm1 +line
	i = i +1
endWhile
jkm1 = jkm1 +"\r" +sep
if changed
	util__stringToFile(fname, jkm1)
	sayUsingVoice(VCTX_Message, mkcs___msgDone, OT_Message)
	scheduleFunction("reloadAllConfigs", 2)
	return 2
else
	sayMessage(OT_Error, mkcs___msgUpdateError)
	return 0
endIf
endFunction

string function mkcs__getCommandKey()
; Returns the key used as a prefix for all command sequences in this application.
; This will be the value passed to the latest call to mkcs__setCommandKey().
return mkcs___commandKey
endFunction

string function mkcs___getActiveKeyboardLayout()
; Returns the name of the keyboard layout in effect or null if this can't be determined.
; Tested in JAWS 9.0.2169.
var string fname, string layout
fname = getJAWSSettingsDirectory() +"\\default.jcf"
layout = iniReadString("Options", "KeyboardType", "", fname)
if !layout
	fname = getSharedSettingsDirectory() +"\\default.jcf"
	layout = iniReadString("Options", "KeyboardType", "", fname)
endIf
return layout
endFunction

string function mkcs___getActiveKeymapFile()
; Returns the full path to the active .jkm file or nul if this can't be determined.
var
	string fname
fname = mkcs___getActiveConfiguration(True)
if !fname
	return ""
endIf
fname = fname +".jkm"
fname = findJAWSSettingsFile(fname)
if !fname
	return ""
endIf
return fname
endFunction

int function mkcs___updateKeyList(string byRef list, string prefix, string jkm, string layout, optional int includeScriptNames)
; Adds any key sequences to the given list that begin with the given prefix.
; Key sequences in the given keyboard layout section of the indicated jkm file are scanned for matches.
; Returns the number (0 or more) of entries added to the list.
var
	string keys,
	string sKey, string sKey0,
	int pos,
	int n, int i,
	int prefixLen,
	int additions
keys = iniReadSectionKeys(layout +" Keys", jkm)
n = stringSegmentCount(keys, "|")
if !keys || !n
	return 0
endIf
prefix = stringStripAllBlanks(prefix)
prefixLen = stringLength(prefix)
additions = 0
i = 1
while i < n
	sKey = stringStripAllBlanks(stringSegment(keys, "|", i))
	sKey0 = sKey
	; Case-insensitive prefix match that doesn't let sKey be a substring of the prefix.
	if stringLeft(sKey, prefixLen) == prefix && stringLength(sKey) > prefixLen
		; +1 cuts off the "&" after the prefix.
		sKey = stringChopLeft(sKey, prefixLen+1)
		; Only consider this key level, not subsequent ones.
		; Also remove the trailing "*" that means "stay at this level" if it is there.
		pos = stringContains(sKey, "&")
		if pos
			sKey = stringLeft(sKey, pos-1)
		elif stringRight(sKey, 1) == "*"
			sKey = stringChopRight(sKey, 1)
		endIf
		; Avoid adding duplicates or keys that are part of this help system.
		if !stringContains(stringLower(sKey), "tab")
		&& !stringContains(stringLower(sKey), "enter")
		&& !stringContains(stringLower(sKey), "esc")
		&& !stringSegmentIndex(list, "|", sKey, True)
			list = list +"|" +sKey
			if includeScriptNames
				list = list +"=" +mkcs___getScriptAssignedTo(sKey0)
			endIf
			additions = additions +1
		endIf
	endIf
	i = i +1
endWhile
return additions
endFunction

string function mkcs___getKeyList(string prefix, optional int includeScriptNames)
; Returns a list of the key sequences starting with the given prefix.
; Key sequences in Common Keys and the active keyboard layout section of the active jkm file are scanned for matches.
; Returned sequences omit the prefix.
var
	string layout,
	string jkm,
	string list
layout = mkcs___getActiveKeyboardLayout()
if !layout
	return ""
endIf
jkm = mkcs___getActiveKeymapFile()
if !jkm
	return ""
endIf
list = ""
mkcs___updateKeyList(list, prefix, jkm, "Common", includeScriptNames)
mkcs___updateKeyList(list, prefix, jkm, layout, includeScriptNames)
list = stringChopLeft(list, 1)  ; remove extra leading separator.
return list
endFunction

const
; Values for iKeyStatus in keymapChangedEvent
	mkcs___KeySequenceComplete = 0,
	mkcs___KeySequenceRepeating = 1,
	mkcs___KeySequencePending = 2,
	mkcs___KeySequenceCanceled = 3,
	mkcs___KeySequenceStart = 4,
	mkcs___KeySequenceRestarted = 5
globals
	int mkcs___active,
	string mkcs___helpText,
	int mkcs___lastKeyCode,
	string mkcs___sequence,
; The next three are used for the Enter/exec functionality.
	string mkcs___lastSeq,
	string mkcs___execScript,
	int mkcs___lastIndex,
	int mkcs___lastKeyStatus,
	int mkcs___helpIndex

int function mkcs___sayKeyPrompt(string seq)
; Speak the prompt for the given (or default) key sequence if there is a prompt for it.
; Returns True if the speaking of the prompt was handled and False if not.
var
	int idx,
	string prompt,
	int isScheduledCall
isScheduledCall = False
if !seq
	seq = mkcs___sequence
	isScheduledCall = True
endIf
; Cut off the command key from the sequence before looking up prompts.
idx = stringContains(seq, "&")
if idx
	seq = stringChopLeft(seq, idx)
else
	seq = ""
endIf
; Find the key's entry in the help text.
idx = stringContains(stringLower(mkcs___helpText), "\10"+stringLower(seq)+"=")
if idx
	; Chop off everything through seq=
	prompt = stringChopLeft(mkcs___helpText, idx+stringLength(seq)+1)
	; Chop off everything past the key's entry.
	prompt = stringSegment(prompt, "\10", 1)
	; Chop off the actual help text (left of the prompt).
	prompt = stringSegment(prompt, "|", 2)
	if prompt
		if !isScheduledCall
			stopSpeech()
		endIf
		sayMessage(OT_Help, prompt)
	endIf
	return True
endIf
return False
endFunction

void function mkcs__setHelpText(string helpText)
; Set the help text to be used by this system.
; Call from AutoStartEvent in the application script file.
mkcs___helpText = "\10" +helpText +"\10"
endFunction

script mkcs__help(string which)
; Implements round-robin keyboard help for the current level of a multikey sequence.
; Which: next to move forward, prev to move backward, cur to stay put.
; These should correspond to Tab, Shift+Tab, and JAWSKey+Tab/Insert+Tab, respectively.
; Exec, which corresponds to Enter, executes the last entry reported.
; For all this to work, appropriate help must have been registered with mkcs__setHelpText().
; Required help format:
;	- One line per help entry.
;	- Line format: <sequence>=<help>[|<prompt>]
;	- <sequence> is the sequence to which the help applies, minus the command prefix and its "&".
;	- <help> is the help text to speak for this sequence. "@" means use the script's jsd Synopsis or, failing that, Description value.
;	- <prompt> (optional) is the prompt to speak when the user actually enters this sequence (not when getting help).
;	- The first line should provide the prompt for the actual command key (e.g., "=|Command").
var
	string prefix,
	int len,
	int pos,
	int i, int n,
	string seg,
	string helps,
	string seq

; Get the prefix, which is the current position in key sequences without the leading command key or its "&".
; Also get len, its length.
; Prefix will end with an "&" if it is not empty.
; It's empty for top-level keys though (keys that immediately follow the command key).
if which == "exec"
	; The sequence should have just now completed,
	; so the sequence is in mkcs___lastSeq, not mkcs___sequence.
	seq = mkcs___lastSeq
else
	seq = mkcs___sequence
endIf
prefix = seq
pos = stringContains(prefix, "&")
if pos
	; Get rid of the leading command character and its "&"
	; Then put an "&" at the end.
	prefix = stringChopLeft(prefix, pos) +"&"
else
	; We're at the top level (right after entry of the command key).
	prefix = ""
endIf
len = stringLength(prefix)

; Loop through help text lines, collecting those that apply to the current level.
; JAWS seems to use \10, not \13, to separate lines in a Messages text string.
n = stringSegmentCount(mkcs___helpText, "\10")
helps = ""
i = 2  ; mkcs__setHelpText prepends a delimiter, so the first segment is null.
while i <= n
	seg = stringSegment(mkcs___helpText, "\10", i)
	seg = stringTrimLeadingBlanks(stringTrimTrailingBlanks(seg))
	; The first check here keeps out the initial Command prompt line.
	; The second part lets through current entries if we're not at the top level.
	; The third lets through top-level entries if that's where we are.
	if stringLeft(seg, 1) != "=" && ((len && stringLeft(seg, stringLength(prefix)) == prefix) || !len)
		; Chop all but the current level's keystroke off this entry if we're not at the top level.
		if len
			seg = stringChopLeft(seg, len)
		endIf
		; Look for entries that are just this level, not deeper ones.
		; This logic allows "&" to occur in help or prompt text.
		if !stringContains(seg, "&") || stringContains(seg, "&") > stringContains(seg, "=")
			; Remove the prompt part if it's there.
			pos = stringContainsFromRight(seg, "|")
			if pos > stringContains(seg, "=")
				seg = stringLeft(seg, pos-1)
			endIf
			helps = helps +"|" +seg
		endIf
	endIf
	i = i +1
endWhile
helps = stringChopLeft(helps, 1)  ; extra leading delimiter

; Figure out which entry to speak, and do it.
if which == "next" || which == "prev"
	if which == "next"
		mkcs___helpIndex = mkcs___helpIndex +1
	elif which == "prev"
		mkcs___helpIndex = mkcs___helpIndex -1
	endIf
	if mkcs___helpIndex <= 0
		mkcs___helpIndex = stringSegmentCount(helps, "|")
	elif mkcs___helpIndex > stringSegmentCount(helps, "|")
		mkcs___helpIndex = 1
	endIf
endIf  ; next or prev
var int idx = mkcs___helpIndex
if which == "exec"
	idx = mkcs___lastIndex
endIf
seg = stringSegment(helps, "|", idx)
if !seg
	; Happens when there are no non-help-system keys defined.
	if which == "exec"
		beep()
	else
		sayMessage(OT_Error, mkcs___msgNoKeys)
	endIf
	return
endIf
if which == "exec"
	var string sKey = stringSegment(seg, "=", 1)
	;sayMessage(OT_User_Requested_Information, sKey)
	; TODO: Disabled as this does not work in a layer: The key is typed
	; but the script is not run. Simulating seq+"&"+sKey actually fails.
	; [DGL, 2014-08-24, JAWS 15.0]
	;simulateKey(sKey)
	if !mkcs___simulateKey(seq, sKey)
		beep()
	endIf
else
	; This says the help text, pauses, then says the keystroke.
	sKey = stringSegment(seg, "=", 1)
	var string txt = stringSegment(seg, "=", 2)
	if txt == "@"
		txt = mkcs___getScriptHelp(mkcs___getScriptAssignedTo(seq+"&"+sKey))
	endIf
	sayMessage(OT_User_Requested_Information, formatString("%2, %1",
		sKey, txt
	))
endIf
endScript

int function mkcs___simulateKey(string seq, string sKey)
; Simulate the final key in a multikey sequence by calling its script.
sKey = formatString("%1&%2", seq, sKey)
var string mapping = mkcs___getScriptAssignedTo(sKey)
if !mapping
	return False
endIf
; Don't call directly because this may cause recursion problems.
; Don't use scheduleFunction directly in case "()" appears in the jkm.
; Schedule a call to handle all this.
mkcs___execScript = mapping
scheduleFunction("mkcs___exec", 0)
return True
endFunction

string function mkcs___getScriptAssignedTo(string sKey)
; Get the script name attached to the given key or sequence.
; Works on sequences where getScriptAssignedTo() fails.
var string layouts = mkcs___getActiveKeyboardLayout()
if layouts
	layouts = layouts +" Keys|"
endIf
layouts = layouts +"Common Keys"
var string files = ""
var string cfg = mkcs___getActiveConfiguration(True)
if cfg
	files = files +formatString("|%1|%2",
		getJAWSSettingsDirectory() +"\\" +cfg +".jkm",
		getSharedSettingsDirectory() +"\\" +cfg +".jkm"
	)
endIf
files = files +formatString("|%1|%2",
	getJAWSSettingsDirectory() +"\\default.jkm",
	getSharedSettingsDirectory() +"\\default.jkm"
)
files = stringChopLeft(files, 1)
var int nfiles = stringSegmentCount(files, "|")
var int nlayouts = stringSegmentCount(layouts, "|")
var int i, int j, string fpath, string layout
var string mapping = ""
for i = 1 to nfiles
	fpath = stringSegment(files, "|", i)
	for j = 1 to nlayouts
		layout = stringSegment(layouts, "|", j)
		mapping = iniReadString(layout, sKey, "", fpath)
		if mapping
			return mapping
		endIf
		mapping = iniReadString(layout, sKey+"*", "", fpath)
		if mapping
			return mapping
		endIf
	endFor
endFor
return ""
endFunction

void function mkcs___exec()
; Execute a script called via Enter.
formatStringWithEmbeddedFunctions("<$"+mkcs___execScript+">")
mkcs___execScript = ""
endFunction

int function mkcs___playExitSound()
; Play the "exit sequence" sound file.
if !mkcs___soundPath
	mkcs___soundPath = mkcs___findSound("mkcsExit.wav")
	if !mkcs___soundPath
		mkcs___soundPath = mkcs___findSound("CameraShutter3.wav")
	endIf
endIf
if mkcs___soundPath
	return playSound(mkcs___soundPath)
endIf
return False
endFunction

string function mkcs___findSound(string sound)
; Find and return the full path to the given sound file,
; or return null if it can't be found.
var string spath, string base
;spath = findJAWSSoundFile(sound)
base = getJAWSSettingsDirectory() +"\\"
spath = base +"sounds\\" +sound
if fileExists(spath)
	return spath
endIf
spath = getJAWSSettingsDirectory() +sound
if fileExists(spath)
	return spath
endIf
base = getSharedSettingsDirectory() +"\\"
spath = base +"sounds\\" +sound
if fileExists(spath)
	return spath
endIf
spath = getJAWSSettingsDirectory() +sound
if fileExists(spath)
	return spath
endIf
return ""
endFunction

void function keymapChangedEvent(int iKeyCode, string sKeyName, int iKeyStatus)
if mkcs__keymapChangedEvent(iKeyCode, sKeyName, iKeyStatus)
	return
endIf
if getJFWVersion() >= 1100000
	; Chaining in older JAWS versions causes "unknown function call to KeyMapChangedEvent" messages for some reason.
	; JAWS 11.0 starts including a KeymapChangedEvent in default.jss.
	keymapChangedEvent(iKeyCode, sKeyName, iKeyStatus)
endIf
endFunction

void function mkcs__keymapChangedEvent(int iKeyCode, string sKeyName, int iKeyStatus)
; Helper for KeyMapChangedEvent.
; Returns True if the default KeyMapChangedEvent should be blocked and False if not.

;sayString(formatString("Last %1 %2, current %4 %5",
;	mkcs___sequence, intToString(mkcs___lastKeyStatus), intToString(mkcs___lastKeyCode),
;sKeyName, intToString(iKeyStatus), intToString(iKeyCode)
;))
var
	string seq
; Filter out unrelated sequences and make any wanted sounds indicating changes into andout of our sequence.
if iKeyStatus == mkcs___KeySequenceStart || iKeyStatus == mkcs___KeySequenceRestarted
	if !isSameKey(sKeyName, mkcs___commandKey)
		; Not our sequence.
		mkcs___active = False
		return False
	endIf
	; Start or restart of our sequence.
	mkcs___active = True
elif !mkcs___active
	; Something other than activation going on, but it's not our sequence.
	return False
elif iKeyStatus == mkcs___KeySequenceCanceled
|| (iKeyStatus == mkcs___KeySequenceComplete && (mkcs___lastKeyStatus != mkcs___KeySequenceComplete && mkcs___lastKeyStatus != mkcs___KeySequenceCanceled))
	; End of our sequence.
	mkcs___playExitSound()
	mkcs___active = False
endIf

; Establish the sequence of keys that got us here.
if iKeyStatus == mkcs___KeySequenceStart || iKeyStatus == mkcs___KeySequenceRestarted
	mkcs___sequence = ""
	mkcs___helpIndex = 0
endIf
seq = ""
if mkcs___sequence
	seq = mkcs___sequence +"&"
endIf
seq = seq +sKeyName

; Announce the new key if there's an announcement to make for it.
if iKeyStatus == mkcs___KeySequenceCanceled
	if iKeyCode == mkcs___lastKeyCode
	&& (mkcs___lastKeyStatus == mkcs___KeySequenceStart || mkcs___lastKeyStatus == mkcs___KeySequenceRestarted)
		; Pass a repeat of the root key through and speak its name.
		sayMessage(OT_Keyboard, sKeyName)
		if !gbKeyboardHelp
			typeKey(sKeyName)
		endIf
	else
		; Avoid saying the invalid key's name and stop any previous speech.
		; Then announce the sequence cancellation.
		stopSpeech()
		;sayMessage(OT_Help, "Canceled", "")
	endIf
elif iKeyStatus != mkcs___KeySequenceComplete || mkcs___lastKeyStatus != mkcs___KeySequenceComplete
	; Say the key sequence's prompt if there is one, or the key name if not..
	; The above condition keeps this block from applying to non-layered script keys.
	if !mkcs___sayKeyPrompt(seq)
		stopSpeech()
		; TODO: Disabled because we don't always want default labels.
		;sayMessage(OT_Keyboard, sKeyName)
	endIf
endIf

; Set the sequence to remember for next time.
; Also set other globals to remember parameters to this call.
if iKeyStatus == mkcs___KeySequenceComplete || iKeyStatus == mkcs___KeySequenceCanceled
	if iKeyStatus == mkcs___KeySequenceComplete
		mkcs___lastSeq = mkcs___sequence
		mkcs___lastIndex = mkcs___helpIndex
	endIf
	mkcs___sequence = ""
	mkcs___helpIndex = 0
elif iKeyStatus == mkcs___KeySequenceRepeating
	; Do nothing; the new key stayed at this level.
	;scheduleFunction("mkcs___sayKeyPrompt", 0)
else
	; Pending:  Include the new key.
	; Start/Restarted:  Start with only the new key (already done above).
	mkcs___sequence = seq
	mkcs___helpIndex = 0
endIf
mkcs___lastKeyCode = iKeyCode
mkcs___lastKeyStatus = iKeyStatus
; This avoids an extra top-level prompt on the first non-Command scripted key after a sequence cancellation.
if iKeyStatus == mkcs___KeySequenceCanceled
	mkcs___lastKeyStatus = mkcs___KeySequenceComplete
endIf

; Chain this event.
;sayString("map " +sKeyName +" " +stringSegment("complete|repeating|pending|canceled|start|restarted", "|", iKeyStatus+1))
return True
endFunction

void function cancelLayeredKeySequence()
if mkcs___active
	mkcs___playExitSound()
	mkcs___active = False
endIf
cancelLayeredKeySequence()
endFunction

script mkcs__exit()
;sayMessage(OT_Help, "Exit", "", False)
endScript

int function mkcs__isSequenceActive()
; Returns True if one of our key sequences is active (in progress).
return mkcs___active
endFunction

string function mkcs___getScriptHelp(string scriptName)
; Get help for the given script. Tries Synopsis, Description, and default Synopsis and Description, in that order.
; On failure, returns the null string.
var string h = getScriptSynopsis(scriptName)
if !h h = getScriptDescription(scriptName) endIf
if !h h = getScriptSynopsis(scriptName, True) endIf
if !h h = getScriptDescription(scriptName, True) endIf
return h
endFunction

;=============================== End module mkcs ===============================


;================================ Module apphelp ===============================
; Automatic access to custom application help (HTM) file.
;
; The help file should be named <configName>.htm, <configName> being the result of getActiveConfiguration().
; Alternatively, the file can be named by a HelpDocument key in a Script Options section of the active configuration's .jcf file.
; As a final alternative, the name can be supplied from autoStartEvent by calling apphelp__setDocName().
;
; Author:  Doug Lee
;
; This is public code.

globals
; Custom doc name if set by a call to apphelp__setDocName().
	string apphelp___docName,
; App in focus when the above name was set.
	string apphelp___docApp,
; Number of quick-succession calls to screenSensitiveHelp seen so far.
; apphelp___doHelp() also sets this to -1 to tell the screenSensitiveHelp script in this file to bypass itself and call up the chain.
	int apphelp___callCount

string function apphelp___getCustomHelpDoc()
; Gets the full path to the custom help document for the running scripts.
; Returns null if no such file is found.
var string docname = ""
var string appname = getActiveConfiguration()
if stringLength(apphelp___docApp)
&& stringCompare(appname, apphelp___docApp, True) == 0
	docname = apphelp___docName
endIf
if !docName
	var string fname = getActiveConfiguration() +".jcf"
	fname = findJAWSSettingsFile(fname)
	if stringLength(fname) && fileExists(fname)
		docname = iniReadString("Script Options", "HelpDocument", "", fname)
	endIf
endIf
if !docname
	docname = getActiveConfiguration()
	if !stringLength(docname) || docname == "default"
		return ""
	endIf
	docname = docname +".htm"
endIf
docname = findJAWSSettingsFile(docname)
if !fileExists(docname)
	return ""
endIf
return docname
endFunction

void function apphelp___doHelp()
; Implements the following behavior based on the number of JAWSKey+F1 presses and the presence or absence of a custom (.htm) help file:
;	One press: Normal ScreenSensitiveHelp in 0.6 seconds.
;	Two quick presses: The custom help document opens if present, or if not, JAWS app-specific (chm) help opens.
;	Three or more quick presses: JAWS-specific (chm) help opens if present.
var
	int callCount,
		string docname
callCount = apphelp___callCount
apphelp___callCount = 0
if callCount != 2
	; 1 = screenSensitiveHelp, 3+ = app-specific help.
	; The standard screenSensitiveHelp script will use isSameScript() to decide what to do.
	apphelp___callCount = -1
	performScript screenSensitiveHelp()
	apphelp___callCount = 0
	return
endIf

docname = apphelp___getCustomHelpDoc()
if docname
	if userBufferIsActive()
		userBufferDeactivate()
	endIf
	sayMessage(OT_Message, apphelp___msgOpening)
	run("\"" +docname +"\"")
	return
endIf

; No custom help document, so revert to standard behavior.
apphelp___callCount = -1
performScript screenSensitiveHelp()
apphelp___callCount = 0
endFunction

script screenSensitiveHelp()
; Intercept that makes the apphelp___doHelp() function take effect.
if apphelp___callCount == -1
	; apphelp___doHelp sets apphelp___callCount to -1 to indicate it wants to call the original screenSensitiveHelp script.
	performScript screenSensitiveHelp()
	return
endIf

apphelp___callCount = apphelp___callCount +1
scheduleFunction("apphelp___doHelp", 6)
endScript

void function apphelp__setDocName(string docname)
; Set the name of the custom help doc to load.
; Call from AutoStartEvent if necessary.
apphelp___docApp = getActiveConfiguration()
apphelp___docname = docname
endFunction

;============================== End module apphelp =============================


;================================ Module string ================================
; String manipulation functions.
;
; Author:  Doug Lee

string function string__segmentNext(string byRef s, string sep)
; Gets, returns, and removes from s the next string segment and the trailing separator if any.
; sep may be multi-character and is case sensitive.
; If sep is null, the entire string is treated as one segment.
var
	int pos,
	string s1
pos = 0
if sep
	pos = stringContains(s, sep)
endIf
if !pos || !sep
	s1 = stringChopLeft(" " +s, 1)  ; avoids pointer duplication
	s = ""
	return s1
endIf
s1 = stringLeft(s, pos-1)
s = stringChopLeft(s, pos +stringLength(sep) -1)
return s1
endFunction

string function string__segmentNextFromRight(string byRef s, string sep)
; Gets, returns, and removes from s the last string segment and the preceding separator if any.
; sep may be multi-character and is case sensitive.
; If sep is null, the entire string is treated as one segment.
var
	int pos,
	string s1
pos = 0
if sep
	pos = stringContainsFromRight(s, sep)
endIf
if !pos || !sep
	s1 = stringChopLeft(" " +s, 1)  ; avoids pointer duplication
	s = ""
	return s1
endIf
s1 = stringChopLeft(s, pos +stringLength(sep) -1)
s = stringLeft(s, pos-1)
return s1
endFunction

int function string__replace(string byRef s, string r1, string r2, string whatToDo)
; Replace r1 in s with r2 if r1 is found where whatToDo indicates.
; Note:  A space is sometimes prepended to and then removed from results in this function to keep JAWS from adding
; numeric strings arithmetically, which can happen in JAWS versions at least through 3.70.47 [DGL, 12/20/2000]
; whatToDo: l=left, r=right, a=anywhere, e=everywhere.
; Add t (for Trim) if whitespace should be removed at the point where s is split.
; Add c (for ConsiderCase) if case-sensitive matching is desired.
; For compatibility with older scripts, WhatToDo can also be an integer sum of flags:
; 0=l, 1=r, 2=a, 3=e, 8=t, 16=c.
; Returns the number of changes made to s, which must be passed by reference.
; If s is null, no changes are made and 0 is returned.
; If r1 is null, the following happens based on WhatToDo:
;	"e" - r2 is inserted between each character pair in s.  All blanks are removed first if "t" is included.
;	"l" or "a" - r2 is inserted after the first character of s, after blanks around it are removed if "t" is included.
;	"r" - r2 is inserted before the last character of s, after blanks around it are removed if "t" is included.
var
	string fromWhere,
	int trimEnds,
	int considerCase,
	int pos,
	int l1,
	string sa,
	string sb,
	int changed,
	int changeCount,
	int done

if !stringLength(s)
	return 0
endIf

; Compatibility with older integer whatToDo parameter.
if stringLength(whatToDo) == 1 && stringContains(whatToDo, "0")
	whatToDo = "l"
	elif stringToInt(whatToDo)
	whatToDo = stringSegment("r|a|e|||||lt|rt|at|et|||||lc|rc|ac|ec|||||ltc|rtc|atc|etc", "|", stringToInt(whatToDo) & 31)
endIf

whatToDo = stringLower(whatToDo)
trimEnds = stringContains(whatToDo, "t")
considerCase = stringContains(whatToDo, "c")
fromWhere = whatToDo

; Special cases for null r1.
if !stringLength(r1) && fromWhere == "e"
	if trimEnds
		sa = stringStripAllBlanks(s)
	else
		sa = string__ref(s)
	endIf
	s = ""
	l1 = stringLength(sa)
	pos = 1
	while pos < l1
		s = string__cat(s, string__cat(substring(sa, pos, 1), r2))
		pos = pos +1
	endWhile
	s = string__cat(s, stringRight(sa, 1))
	return stringLength(sa)
elif !stringLength(r1)
	if fromWhere == "l" || fromWhere == "a"
		if trimEnds
			sb = stringTrimLeadingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringLeft(sb, 1)
		sb = stringChopLeft(sb, 1)
		if trimEnds
			sb = stringTrimLeadingBlanks(sb)
		endIf
	else  ; fromWhere == "r"
		if trimEnds
			sb = stringTrimTrailingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringChopRight(sb, 1)
		if trimEnds
			sa = stringTrimTrailingBlanks(sa)
		endIf
		sb = stringRight(sb, 1)
	endIf
	if !stringLength(sa) || !stringLength(sb)
		return 0
	endIf
	s = string__cat(sa, string__cat(r2, sb))
	return 1
endIf

changeCount = 0
while !done
	l1 = stringLength(r1)
	changed = false
	if fromWhere == "l"
		if !stringCompare(stringLeft(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimLeadingBlanks(stringChopLeft(s, l1))
				s = r2 + s
			else
				s = r2 + stringChopLeft(s, l1)
			endIf
			changed = true
		endIf
	elif fromWhere == "r"
		if !stringCompare(stringRight(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimTrailingBlanks(stringChopRight(s, l1))
				s = s + r2
			else
				s = stringChopRight(s, l1) + r2
			endIf
			changed = true
		endIf
	else  ; fromWhere == "a" or "e"
		if considerCase
			pos = StringContains(s, r1)
		else
			pos = StringContains(stringLower(s), stringLower(r1))
		endIf
		if pos
			if trimEnds
				sa = stringTrimTrailingBlanks(stringLeft(s, pos-1))
				sb = stringTrimLeadingBlanks(stringChopLeft(s, pos+l1-1))
				s = " " +sa +r2 +sb
				s = stringChopLeft(s, 1)
			else
				s = " " +stringLeft(s, pos-1) +r2 +stringChopLeft(s, pos+l1-1)
				s = stringChopLeft(s, 1)
			endIf
			changed = True
		endIf
	endIf
	if fromWhere == "e"
		done = !changed
	else
		done = True
	endIf
	if changed
		changeCount = changeCount +1
	endIf
endWhile
return changeCount
endFunction

string function string__ref(variant v)
; Type cast from anything to string.
; Useful for allowing object results to be included in string concatenations without a compile-time error.
return v
endFunction

string function string__cat(string s1, string s2)
; Safe string concatenator.
return stringChopLeft(" " +s1 +s2, 1)
endFunction

;============================== End module string ==============================


;================================= Module util =================================
; Miscellaneous utility functions
;
; Author:  Doug Lee

string function util__stringFromFile(string fname, /*?*/ string fmt)
; Returns the file contents as a string.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt,
	string s
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 1, 0, iFmt)  ; 1 is ForReading, 0 means don't create
if !f
	;sayString("No code file")
	return ""
endIf
s = f.readAll()
f.close()
if !stringLength(s)
	;sayString("No code in the file")
endIf
return s
endFunction

int function util__stringToFile(string fname, string s, /*?*/ string fmt)
; Writes the string to the indicated file, creating or replacing the file.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 2, 1, iFmt)  ; 2 is ForWriting, 1 means create
if !f
	;sayString("Can't create file")
	return ""
endIf
f.write(s)
f.close()
return True
endFunction

object function util__XMLDocument(string xml, optional int silent, int preserveWhitespace)
; Return an MSXML DOM document object for the given XML. Returns null on error.
; If silent is not passed and True, error messages speak as appropriate.
; This includes parse errors from oDoc.loadXML().
; Whitespace is preserved if preserveWhitespace is passed and True, otherwise not.
if !xml
	return
endIf
var object oDoc = util__createObject("msxml2.DOMDocument.6.0")
if !oDoc
	return
endIf
oDoc.async = False
oDoc.resolveExternals = False
oDoc.preserveWhiteSpace = preserveWhitespace
oDoc.loadXML(xml)
if oDoc.parseError.errorCode
	if !silent
		sayMessage(OT_Error, formatString("%1 at line %2 column %3: %4",
			oDoc.parseError.reason, oDoc.parseError.line, oDoc.parseError.linepos,
			oDoc.parseError.srcText
		))
	endIf
	return
endIf
return oDoc
endFunction

object function util__createObject(string objname)
; A CreateObject wrapper that avoids causing 5-7 second delays in DOS boxes when called if JAWS is running as a service.
; As of June 11, 2011, uses the CoCreateInstance method of creating an object
; unless that doesn't work, whereupon the GetObjectHelper method is used.
; Previously the CoCreateInstance method was only used if JAWS was running as a service.
var object o = createObjectEx(objname, False)
if !o
	o = createObject(objname)
endIf
return o
endFunction

;=============================== End module util ===============================

